# -*- coding: utf-8 -*-

from __future__ import absolute_import

__version__ = "1.0.1"

# the public api
from homebase.homebase import (
    user_data_dir, user_cache_dir, user_config_dir,
    user_logs_dir, user_state_dir, site_data_dir, site_config_dir, site_data_dir_list, site_config_dir_list
)



# -*- coding: utf-8 -*-

import re
from setuptools import setup, find_packages
from os.path import abspath, join, dirname, realpath

here = abspath(dirname(realpath(__file__)))


def get_version(folder):
    with open(join(folder, 'homebase', '__init__.py')) as init_file:
        regex = """^__version__ = ['"]([^'"]*)['"]"""
        version_match = re.search(regex, init_file.read(), re.M)  # using re.M, ^ is matched to beginning of each line.
        if not version_match:
            raise RuntimeError("Unable to find a version")
        return version_match.group(1)


def readme(folder):
    with open(join(folder, 'README.md')) as desc_file:
        return desc_file.read()


def requirements(folder):
    try:
        with open(join(folder, 'requirements.txt')) as req_file:
            return req_file.readlines()
    except IOError:
        return []


version = get_version(here)

setup(
    name='homebase',
    version=version,
    description='A platform independent API for paths in which applications can write data.',
    author='D-Wave Systems Inc.',
    author_email='oshklarsky@dwavesys.com',
    long_description=readme(here),
    packages=find_packages(),
    install_requires=requirements(here),
    license='Apache 2.0',
    url='https://github.com/dwavesystems/homebase',
    download_url='https://github.com/dwavesystems/homebase/archive/{}.tar.gz'.format(version)
)



"""
Test homebase.

Since some of the homebase functionality is OS specific, this module can only get 100% coverage if it is run on
all three operating systems: Windows, Linux, and macOS. Thus, coverage tools will need to merge reports from different
runs.

From a functional perspective, the only thing that changes across operating systems are expected results.
As such, the strategy for testing is as follows:

1. A base class TestHomebase defines basic expectations and all of the unit tests common to all three operating
   systems.
2. A class for handling virtualenv tests, TestHomebaseVirtualEnv that modifies the expected path for each
   OS to expect the appropriate virtualenv base dir, and runs all tests of TestHomebase by inheriting from it.
3. The base class TestHomebase tests linux in the case where the XDG variables are not set. Another class
   TestHomebaseLinuxXDG sets these variables and the expected paths and reruns the tests by inheriting from
   TestHomebase.
4.A class for Windows TestHomebaseWindows adds tests for roaming and app_author values.

To run the tests against linux, you can use the accompanying Dockerfile.

"""
import os
import sys
import shutil
import unittest

import virtualenv
import homebase

_PY2 = sys.version_info[0] == 2
if _PY2:
    def itervalues(d):
        return d.itervalues()
else:
    def itervalues(d):
        return iter(d.values())


class TestHomebase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.app_name = "test_app"
        self.app_author = 'for_windows_only'

        # we can't just use sys.platform because on linux it could be either linux2 or just linux
        # (https://docs.python.org/2/library/sys.html#sys.platform)
        # If we're masking linux, might as well mask darwin with mac os.
        if sys.platform.startswith('linux'):
            self.platform = 'linux'
        elif sys.platform == 'darwin':
            self.platform = 'mac_os'
        else:
            self.platform = 'windows'

        self.base_paths = self._expected_base_paths()

    def _expected_base_paths(self):
        mac_os_app_support = os.path.expanduser('~/Library/Application Support/')
        mac_os_site_app_support = '/Library/Application Support'
        base_paths = {
            'mac_os': {
                'user_data': mac_os_app_support,
                'user_config': mac_os_app_support,
                'user_state': mac_os_app_support,
                'user_cache': os.path.expanduser('~/Library/Caches'),
                'user_log': os.path.expanduser('~/Library/Logs'),
                'site_data': [mac_os_site_app_support],
                'site_config': [mac_os_site_app_support]
            },
            'linux': {
                'user_data': os.path.expanduser("~/.local/share"),
                'user_config': os.path.expanduser('~/.config'),
                'user_state': os.path.expanduser('~/.local/state'),
                'user_cache': os.path.expanduser('~/.cache'),
                'user_log': os.path.expanduser('~/.log'),
                'site_data': ['/usr/local/share', '/usr/share'],
                'site_config': ['/etc/xdg']
            },
            'windows': self._windows_base_paths()
        }

        # add virtualenv expectations. When there is no actual virtualenv, we expect the use_virtualenv parameter
        # to do nothing.
        for paths in itervalues(base_paths):
            paths['user_data_venv'] = paths['user_data']
            paths['user_config_venv'] = paths['user_config']
            paths['user_state_venv'] = paths['user_state']
            paths['user_cache_venv'] = paths['user_cache']
            paths['user_log_venv'] = paths['user_log']

        return base_paths

    def _windows_base_paths(self):
        windows_username = os.getenv('username')
        base_paths = {
            'user_data': '',
            'user_config': '',
            'user_state': '',
            'user_cache': '',
            'user_log': '',
            'site_data': '',
            'site_config': ''

        }
        if windows_username is not None:
            windows_absolute_path = os.path.join('C:', os.sep)
            windows_base_path = os.path.join(windows_absolute_path, 'Users', windows_username, 'AppData', 'Local')

            # add windows expectations.
            base_paths['user_data'] = os.path.join(windows_base_path, self.app_author)
            base_paths['user_config'] = os.path.join(windows_base_path, self.app_author)
            base_paths['user_state'] = os.path.join(windows_base_path, self.app_author)
            base_paths['user_cache'] = os.path.join(windows_base_path, self.app_author, 'Caches')
            base_paths['user_log'] = os.path.join(windows_base_path, self.app_author, 'Logs')
            base_paths['site_data'] = [os.path.join(windows_absolute_path, 'ProgramData', self.app_author)]
            base_paths['site_config'] = [os.path.join(windows_absolute_path, 'ProgramData', self.app_author)]

        return base_paths

    @staticmethod
    def _setup_linux_xdg_vars():
        test_dir = os.path.abspath(os.path.dirname(__file__))
        os.environ['XDG_DATA_HOME'] = os.path.join(test_dir, '.local/share')
        os.environ['XDG_CONFIG_HOME'] = os.path.join(test_dir, '.config')
        os.environ['XDG_STATE_HOME'] = os.path.join(test_dir, '.local/state')
        os.environ['XDG_CACHE_HOME'] = os.path.join(test_dir, '.cache')
        os.environ['XDG_DATA_DIRS'] = os.pathsep.join([os.path.join(test_dir, '.site/data'),
                                                       os.path.join(test_dir, '.site/data2')])
        os.environ['XDG_CONFIG_DIRS'] = os.pathsep.join([os.path.join(test_dir, '.site/config'),
                                                         os.path.join(test_dir, '.site/config2')])

    @staticmethod
    def _clear_linux_xdg_vars():
        var_names = ['XDG_DATA_HOME', 'XDG_CONFIG_HOME', 'XDG_STATE_HOME', 'XDG_CACHE_HOME', 'XDG_DATA_DIRS',
                     'XDG_CONFIG_DIRS']
        for var in var_names:
            if var in os.environ:
                del os.environ[var]

    @classmethod
    def setUpClass(cls):
        cls._clear_linux_xdg_vars()

    #####################################################
    # user_data_dir.
    #####################################################
    def test_user_data_no_version_no_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_data'], self.app_name)
        self.assertEqual(expected,
                         homebase.user_data_dir(self.app_name, app_author=self.app_author, version=None,
                                                use_virtualenv=False, create=False))

    def test_user_data_no_version_no_venv_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_data'], self.app_name)
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertEqual(expected, homebase.user_data_dir(self.app_name, app_author=self.app_author, version=None,
                                                          use_virtualenv=False, create=True))
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_data_no_version_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_data_venv'], self.app_name)
        self.assertEqual(expected, homebase.user_data_dir(self.app_name, app_author=self.app_author, version=None,
                                                          use_virtualenv=True, create=False))

    def test_user_data_no_version_venv_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_data_venv'], self.app_name)
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertEqual(expected, homebase.user_data_dir(self.app_name, app_author=self.app_author, version=None,
                                                          use_virtualenv=True, create=True))
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_data_version_no_venv_no_create(self):
        version = "1.0"
        expected = os.path.join(self.base_paths[self.platform]['user_data'], '{}_{}'.format(self.app_name, version))
        self.assertEqual(expected,
                         homebase.user_data_dir(self.app_name, app_author=self.app_author, version=version,
                                                use_virtualenv=False, create=False))

    def test_user_data_version_no_venv_create(self):
        version = "1.0"
        expected = os.path.join(self.base_paths[self.platform]['user_data'], '{}_{}'.format(self.app_name, version))
        if os.path.exists(expected):
            shutil.rmtree(expected)
        result = homebase.user_data_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=False,
                                        create=True)
        self.assertEqual(expected, result)
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_data_version_venv_no_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['user_data_venv'],
                                '{}_{}'.format(self.app_name, version))
        result = homebase.user_data_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=True,
                                        create=False)
        self.assertEqual(expected, result)

    def test_user_data_version_venv_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['user_data_venv'],
                                '{}_{}'.format(self.app_name, version))
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertFalse(os.path.exists(expected))
        result = homebase.user_data_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=True,
                                        create=True)
        self.assertEqual(expected, result)
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    #####################################################
    # user_cache_dir.
    #####################################################
    def test_user_cache_no_version_no_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_cache'], self.app_name)
        self.assertEqual(expected,
                         homebase.user_cache_dir(self.app_name, app_author=self.app_author, version=None,
                                                 use_virtualenv=False, create=False))

    def test_user_cache_no_version_no_venv_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_cache'], self.app_name)
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertEqual(expected,
                         homebase.user_cache_dir(self.app_name, app_author=self.app_author, version=None,
                                                 use_virtualenv=False, create=True))
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_cache_no_version_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_cache_venv'], self.app_name)
        self.assertEqual(expected,
                         homebase.user_cache_dir(self.app_name, app_author=self.app_author, version=None,
                                                 use_virtualenv=True, create=False))

    def test_user_cache_no_version_venv_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_cache_venv'], self.app_name)
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertEqual(expected, homebase.user_cache_dir(self.app_name, app_author=self.app_author, version=None,
                                                           use_virtualenv=True, create=True))
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_cache_version_no_venv_no_create(self):
        version = "1.0"
        expected = os.path.join(self.base_paths[self.platform]['user_cache'], '{}_{}'.format(self.app_name, version))
        self.assertEqual(expected,
                         homebase.user_cache_dir(self.app_name, app_author=self.app_author, version=version,
                                                 use_virtualenv=False, create=False))

    def test_user_cache_version_no_venv_create(self):
        version = "1.0"
        expected = os.path.join(self.base_paths[self.platform]['user_cache'], '{}_{}'.format(self.app_name, version))
        if os.path.exists(expected):
            shutil.rmtree(expected)
        result = homebase.user_cache_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=False,
                                         create=True)
        self.assertEqual(expected, result)
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_cache_version_venv_no_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['user_cache_venv'],
                                '{}_{}'.format(self.app_name, version))
        result = homebase.user_cache_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=True,
                                         create=False)
        self.assertEqual(expected, result)

    def test_user_cache_version_venv_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['user_cache_venv'],
                                '{}_{}'.format(self.app_name, version))
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertFalse(os.path.exists(expected))
        result = homebase.user_cache_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=True,
                                         create=True)
        self.assertEqual(expected, result)
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    #####################################################
    # user_config_dir
    #####################################################
    def test_user_config_no_version_no_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_config'], self.app_name)
        self.assertEqual(expected,
                         homebase.user_config_dir(self.app_name, app_author=self.app_author, version=None,
                                                  use_virtualenv=False, create=False))

    def test_user_config_no_version_no_venv_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_config'], self.app_name)
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertEqual(expected,
                         homebase.user_config_dir(self.app_name, app_author=self.app_author, version=None,
                                                  use_virtualenv=False, create=True))
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_config_no_version_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_config_venv'], self.app_name)
        self.assertEqual(expected,
                         homebase.user_config_dir(self.app_name, app_author=self.app_author, version=None,
                                                  use_virtualenv=True, create=False))

    def test_user_config_no_version_venv_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_config_venv'], self.app_name)
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertEqual(expected,
                         homebase.user_config_dir(self.app_name, app_author=self.app_author, version=None,
                                                  use_virtualenv=True, create=True))
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_config_version_no_venv_no_create(self):
        version = "1.0"
        expected = os.path.join(self.base_paths[self.platform]['user_config'], '{}_{}'.format(self.app_name, version))
        self.assertEqual(expected,
                         homebase.user_config_dir(self.app_name, app_author=self.app_author, version=version,
                                                  use_virtualenv=False, create=False))

    def test_user_config_version_no_venv_create(self):
        version = "1.0"
        expected = os.path.join(self.base_paths[self.platform]['user_config'], '{}_{}'.format(self.app_name, version))
        if os.path.exists(expected):
            shutil.rmtree(expected)
        result = homebase.user_config_dir(self.app_name, app_author=self.app_author, version=version,
                                          use_virtualenv=False, create=True)
        self.assertEqual(expected, result)
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_config_version_venv_no_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['user_config_venv'],
                                '{}_{}'.format(self.app_name, version))
        result = homebase.user_config_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=True,
                                          create=False)
        self.assertEqual(expected, result)

    def test_user_config_version_venv_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['user_config_venv'],
                                '{}_{}'.format(self.app_name, version))
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertFalse(os.path.exists(expected))
        result = homebase.user_config_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=True,
                                          create=True)
        self.assertEqual(expected, result)
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    #####################################################
    # user_state_dir
    #####################################################
    def test_user_state_no_version_no_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_state'], self.app_name)
        self.assertEqual(expected,
                         homebase.user_state_dir(self.app_name, app_author=self.app_author, version=None,
                                                 use_virtualenv=False, create=False))

    def test_user_state_no_version_no_venv_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_state'], self.app_name)
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertEqual(expected,
                         homebase.user_state_dir(self.app_name, app_author=self.app_author, version=None,
                                                 use_virtualenv=False, create=True))
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_state_no_version_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_state_venv'], self.app_name)
        self.assertEqual(expected,
                         homebase.user_state_dir(self.app_name, app_author=self.app_author, version=None,
                                                 use_virtualenv=True, create=False))

    def test_user_state_no_version_venv_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_state_venv'], self.app_name)
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertEqual(expected, homebase.user_state_dir(self.app_name, app_author=self.app_author, version=None,
                                                           use_virtualenv=True, create=True))
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_state_version_no_venv_no_create(self):
        version = "1.0"
        expected = os.path.join(self.base_paths[self.platform]['user_state'], '{}_{}'.format(self.app_name, version))
        self.assertEqual(expected,
                         homebase.user_state_dir(self.app_name, app_author=self.app_author, version=version,
                                                 use_virtualenv=False, create=False))

    def test_user_state_version_no_venv_create(self):
        version = "1.0"
        expected = os.path.join(self.base_paths[self.platform]['user_state'], '{}_{}'.format(self.app_name, version))
        if os.path.exists(expected):
            shutil.rmtree(expected)
        result = homebase.user_state_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=False,
                                         create=True)
        self.assertEqual(expected, result)
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_state_version_venv_no_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['user_state_venv'],
                                '{}_{}'.format(self.app_name, version))
        result = homebase.user_state_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=True,
                                         create=False)
        self.assertEqual(expected, result)

    def test_user_state_version_venv_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['user_state_venv'],
                                '{}_{}'.format(self.app_name, version))
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertFalse(os.path.exists(expected))
        result = homebase.user_state_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=True,
                                         create=True)
        self.assertEqual(expected, result)
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    #####################################################
    # user_logs_dir
    #####################################################
    def test_user_log_no_version_no_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_log'], self.app_name)
        self.assertEqual(expected, homebase.user_logs_dir(self.app_name, app_author=self.app_author, version=None,
                                                          use_virtualenv=False, create=False))

    def test_user_log_no_version_no_venv_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_log'], self.app_name)
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertEqual(expected, homebase.user_logs_dir(self.app_name, app_author=self.app_author, version=None,
                                                          use_virtualenv=False, create=True))
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_log_no_version_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_log_venv'], self.app_name)
        self.assertEqual(expected, homebase.user_logs_dir(self.app_name, app_author=self.app_author, version=None,
                                                          use_virtualenv=True, create=False))

    def test_user_log_no_version_venv_create(self):
        expected = os.path.join(self.base_paths[self.platform]['user_log_venv'], self.app_name)
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertEqual(expected, homebase.user_logs_dir(self.app_name, app_author=self.app_author, version=None,
                                                          use_virtualenv=True, create=True))
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_log_version_no_venv_no_create(self):
        version = "1.0"
        expected = os.path.join(self.base_paths[self.platform]['user_log'], '{}_{}'.format(self.app_name, version))
        self.assertEqual(expected,
                         homebase.user_logs_dir(self.app_name, app_author=self.app_author, version=version,
                                                use_virtualenv=False, create=False))

    def test_user_log_version_no_venv_create(self):
        version = "1.0"
        expected = os.path.join(self.base_paths[self.platform]['user_log'], '{}_{}'.format(self.app_name, version))
        if os.path.exists(expected):
            shutil.rmtree(expected)
        result = homebase.user_logs_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=False,
                                        create=True)
        self.assertEqual(expected, result)
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    def test_user_log_version_venv_no_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['user_log_venv'],
                                '{}_{}'.format(self.app_name, version))
        result = homebase.user_logs_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=True,
                                        create=False)
        self.assertEqual(expected, result)

    def test_user_log_version_venv_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['user_log_venv'],
                                '{}_{}'.format(self.app_name, version))
        if os.path.exists(expected):
            shutil.rmtree(expected)
        self.assertFalse(os.path.exists(expected))
        result = homebase.user_logs_dir(self.app_name, app_author=self.app_author, version=version, use_virtualenv=True,
                                        create=True)
        self.assertEqual(expected, result)
        self.assertTrue(os.path.exists(expected))
        shutil.rmtree(expected)

    #####################################################
    # site_data_homebase. Not testing create=True since we don't know if we have permissions.
    #####################################################
    def test_site_data_no_version_no_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['site_data'][0], self.app_name)
        self.assertEqual(expected,
                         homebase.site_data_dir(self.app_name, app_author=self.app_author, version=None,
                                                use_virtualenv=False, create=False))

    def test_site_data_no_version_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['site_data'][0], self.app_name)
        self.assertEqual(expected,
                         homebase.site_data_dir(self.app_name, app_author=self.app_author, version=None,
                                                use_virtualenv=True, create=False))

    def test_site_data_version_no_venv_no_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['site_data'][0], '{}_{}'.format(self.app_name, version))
        self.assertEqual(expected,
                         homebase.site_data_dir(self.app_name, app_author=self.app_author, version=version,
                                                use_virtualenv=False, create=False))

    def test_site_data_version_venv_no_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['site_data'][0], '{}_{}'.format(self.app_name, version))
        self.assertEqual(expected,
                         homebase.site_data_dir(self.app_name, app_author=self.app_author, version=version,
                                                use_virtualenv=True, create=False))

    def test_site_data_list_no_version_no_venv_no_create(self):
        expected = [os.path.join(d, self.app_name) for d in self.base_paths[self.platform]['site_data']]
        self.assertEqual(expected,
                         homebase.site_data_dir_list(self.app_name, app_author=self.app_author, version=None,
                                                     use_virtualenv=False, create=False))

    def test_site_data_list_no_version_venv_no_create(self):
        expected = [os.path.join(d, self.app_name) for d in self.base_paths[self.platform]['site_data']]
        self.assertEqual(expected,
                         homebase.site_data_dir_list(self.app_name, app_author=self.app_author, version=None,
                                                     use_virtualenv=True, create=False))

    def test_site_data_list_version_no_venv_no_create(self):
        version = "1.0"
        expected = [os.path.join(d, '{}_{}'.format(self.app_name, version))
                    for d in self.base_paths[self.platform]['site_data']]
        self.assertEqual(expected,
                         homebase.site_data_dir_list(self.app_name, app_author=self.app_author, version=version,
                                                     use_virtualenv=False, create=False))

    def test_site_data_list_version_venv_no_create(self):
        version = '1.0'
        expected = [os.path.join(d, '{}_{}'.format(self.app_name, version))
                    for d in self.base_paths[self.platform]['site_data']]
        result = homebase.site_data_dir_list(self.app_name, app_author=self.app_author, version=version,
                                             use_virtualenv=True, create=False)
        self.assertEqual(expected, result)

    #####################################################
    # site_config_homebase. Not testing create=True since we don't know if we have permissions.
    #####################################################
    def test_site_config_no_version_no_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['site_config'][0], self.app_name)
        self.assertEqual(expected,
                         homebase.site_config_dir(self.app_name, app_author=self.app_author, version=None,
                                                  use_virtualenv=False, create=False))

    def test_site_config_no_version_venv_no_create(self):
        expected = os.path.join(self.base_paths[self.platform]['site_config'][0], self.app_name)
        self.assertEqual(expected,
                         homebase.site_config_dir(self.app_name, app_author=self.app_author, version=None,
                                                  use_virtualenv=True, create=False))

    def test_site_config_version_no_venv_no_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['site_config'][0],
                                '{}_{}'.format(self.app_name, version))
        self.assertEqual(expected,
                         homebase.site_config_dir(self.app_name, app_author=self.app_author, version=version,
                                                  use_virtualenv=False, create=False))

    def test_site_config_version_venv_no_create(self):
        version = '1.0'
        expected = os.path.join(self.base_paths[self.platform]['site_config'][0],
                                '{}_{}'.format(self.app_name, version))
        self.assertEqual(expected,
                         homebase.site_config_dir(self.app_name, app_author=self.app_author, version=version,
                                                  use_virtualenv=True, create=False))

    def test_site_config_list_no_version_no_venv_no_create(self):
        expected = [os.path.join(d, self.app_name) for d in self.base_paths[self.platform]['site_config']]
        self.assertEqual(expected,
                         homebase.site_config_dir_list(self.app_name, app_author=self.app_author, version=None,
                                                       use_virtualenv=False, create=False))

    def test_site_config_list_no_version_venv_no_create(self):
        expected = [os.path.join(d, self.app_name) for d in self.base_paths[self.platform]['site_config']]
        self.assertEqual(expected,
                         homebase.site_config_dir_list(self.app_name, app_author=self.app_author, version=None,
                                                       use_virtualenv=True, create=False))

    def test_site_config_list_version_no_venv_no_create(self):
        version = "1.0"
        expected = [os.path.join(d, '{}_{}'.format(self.app_name, version))
                    for d in self.base_paths[self.platform]['site_config']]
        self.assertEqual(expected,
                         homebase.site_config_dir_list(self.app_name, app_author=self.app_author, version=version,
                                                       use_virtualenv=False, create=False))

    def test_site_config_list_version_venv_no_create(self):
        version = '1.0'
        expected = [os.path.join(d, '{}_{}'.format(self.app_name, version))
                    for d in self.base_paths[self.platform]['site_config']]
        result = homebase.site_config_dir_list(self.app_name, app_author=self.app_author, version=version,
                                               use_virtualenv=True, create=False)
        self.assertEqual(expected, result)


@unittest.skipIf('travis' in os.path.expanduser('~'), 'Skipping travis virtualenv tests for now.')
class TestHomebaseVirtualEnv(TestHomebase):

    virtualenv_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'test_venv')
    real_prefix = sys.prefix

    def __init__(self, *args, **kwargs):
        TestHomebase.__init__(self, *args, **kwargs)

    def _expected_base_paths(self):
        base_paths = TestHomebase._expected_base_paths(self)

        # add virtualenv expectations.
        for paths in itervalues(base_paths):
            paths['user_data_venv'] = os.path.join(self.virtualenv_dir, 'data')
            paths['user_config_venv'] = os.path.join(self.virtualenv_dir, 'config')
            paths['user_state_venv'] = os.path.join(self.virtualenv_dir, 'state')
            paths['user_cache_venv'] = os.path.join(self.virtualenv_dir, 'cache')
            paths['user_log_venv'] = os.path.join(self.virtualenv_dir, 'logs')

        return base_paths

    @classmethod
    def setUpClass(cls):
        virtualenv.create_environment(cls.virtualenv_dir)
        if sys.platform == 'win32':
            activate_script = os.path.join(cls.virtualenv_dir, 'Scripts', 'activate_this.py')
        else:
            activate_script = os.path.join(cls.virtualenv_dir, 'bin', 'activate_this.py')

        if sys.version_info.major == '2':
            execfile(activate_script, dict(__file__=activate_script))
        else:
            with open(activate_script) as activate:
                exec(activate.read(), dict(__file__=activate_script))

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.virtualenv_dir)
        if hasattr(sys, 'real_prefix'):
            sys.prefix = sys.real_prefix
            delattr(sys, 'real_prefix')
        else:
            sys.prefix = sys.base_prefix


@unittest.skipUnless(sys.platform.startswith('linux'), 'TestHomebaseLinuxXDG: Not Linux')
@unittest.skipIf('travis' in os.path.expanduser('~'), 'Skipping travis virtualenv tests for now.')
class TestHomebaseLinuxXDG(TestHomebase):

    def setUp(self):
        self._setup_linux_xdg_vars()

    def tearDown(self):
        self._clear_linux_xdg_vars()

    def _expected_base_paths(self):
        """
        Note: This class will always run regardless of OS, but unless self.platform is linux the values below are
        ignored and the test will be meaningless. You can either run it on linux (e.g. using the Dockerfile at the root
        of this project),or on mac by hard coding self.platform to 'linux' e.g. by overriding __init__.
        """
        base_paths = TestHomebase._expected_base_paths(self)

        self._setup_linux_xdg_vars()

        base_paths['linux']['user_data'] = os.environ['XDG_DATA_HOME']
        base_paths['linux']['user_data_venv'] = os.environ['XDG_DATA_HOME']
        base_paths['linux']['user_config'] = os.environ['XDG_CONFIG_HOME']
        base_paths['linux']['user_config_venv'] = os.environ['XDG_CONFIG_HOME']
        base_paths['linux']['user_state'] = os.environ['XDG_STATE_HOME']
        base_paths['linux']['user_state_venv'] = os.environ['XDG_STATE_HOME']
        base_paths['linux']['user_cache'] = os.environ['XDG_CACHE_HOME']
        base_paths['linux']['user_cache_venv'] = os.environ['XDG_CACHE_HOME']
        base_paths['linux']['site_data'] = os.environ['XDG_DATA_DIRS'].split(os.pathsep)
        base_paths['linux']['site_config'] = os.environ['XDG_CONFIG_DIRS'].split(os.pathsep)

        return base_paths


@unittest.skipUnless(sys.platform == 'win32', 'TestHomebaseWindowsExtra: Not Windows')
class TestHomebaseWindowsExtra(TestHomebase):

    def setUp(self):
        windows_username = os.getenv('username')
        windows_absolute_path = os.path.join('C:', os.sep)
        self.roaming_base_path = os.path.join(windows_absolute_path, 'Users', windows_username, 'AppData', 'Roaming')

        # # add windows expectations.
        # base_paths['user_data'] = os.path.join(windows_base_path, self.app_author)

    def test_user_data_no_app_author(self):
        with self.assertRaises(RuntimeError):
            homebase.user_data_dir(self.app_name, app_author=None)

    def test_user_config_no_app_author(self):
        with self.assertRaises(RuntimeError):
            homebase.user_config_dir(self.app_name, app_author=None)

    def test_user_cache_no_app_author(self):
        with self.assertRaises(RuntimeError):
            homebase.user_cache_dir(self.app_name, app_author=None)

    def test_user_logs_no_app_author(self):
        with self.assertRaises(RuntimeError):
            homebase.user_logs_dir(self.app_name, app_author=None)

    def test_user_state_no_app_author(self):
        with self.assertRaises(RuntimeError):
            homebase.user_state_dir(self.app_name, app_author=None)

    def test_site_data_no_app_author(self):
        with self.assertRaises(RuntimeError):
            homebase.site_data_dir(self.app_name, app_author=None)

    def test_site_config_no_app_author(self):
        with self.assertRaises(RuntimeError):
            homebase.site_config_dir(self.app_name, app_author=None)

    def test_user_data_roaming(self):
        expected = os.path.join(self.roaming_base_path, self.app_author, self.app_name)
        self.assertEqual(expected, homebase.user_data_dir(self.app_name, self.app_author, roaming=True))

    def test_user_config_roaming(self):
        expected = os.path.join(self.roaming_base_path, self.app_author, self.app_name)
        self.assertEqual(expected, homebase.user_config_dir(self.app_name, self.app_author, roaming=True))

    def test_user_state_roaming(self):
        expected = os.path.join(self.roaming_base_path, self.app_author, self.app_name)
        self.assertEqual(expected, homebase.user_state_dir(self.app_name, self.app_author, roaming=True))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides an API for determining application specific directories for data, config, logs, etc.

public module members:

.. autosummary::

    user_data_dir
    user_config_dir
    user_cache_dir
    user_state_dir
    user_logs_dir
    site_data_dir
    site_config_dir
    site_data_dir_list
    site_config_dir_list

This code is inspired by and builds on top of code from http://github.com/ActiveState/appdirs

"""
# Dev Notes:
# - MSDN on where to store app data files:
#   http://support.microsoft.com/default.aspx?scid=kb;en-us;310294#XSLTH3194121123120121120120
# - Mac OS X: http://developer.apple.com/documentation/MacOSX/Conceptual/BPFileSystem/index.html
# - XDG spec for Un*x: http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html

import sys
import os
from enum import Enum


class _FolderTypes(Enum):
    data = 1
    config = 2
    cache = 3
    state = 4
    logs = 5


class Platform(Enum):
    WINDOWS = 1
    MACOS = 2
    POSIX = 3


# only calculate this once, at import.
if sys.platform == 'win32':
    platform = Platform.WINDOWS
elif sys.platform == 'darwin':
    platform = Platform.MACOS
else:
    platform = Platform.POSIX


def user_data_dir(app_name, app_author, version=None, roaming=False, use_virtualenv=True, create=True):
    """
    Return the full path to the user data dir for this application, using a virtualenv location as a base, if it is
    exists, and falling back to the host OS's convention if it doesn't.

    If using a virtualenv, the path returned is :bash:`/path/to/virtualenv/data/app_name`

    Typical user data directories are:

        * Mac OS X:               :bash:`~/Library/Application Support/<app_name>`
        * Unix:                   :bash:`~/.local/share/<app_name>    # or $XDG_DATA_HOME/<app_name>, if defined.`
        * Win XP (not roaming):   :bash:`C:\\Documents and Settings\\<username>\\Application Data\\<app_author>\\<app_name>`
        * Win XP (roaming):       :bash:`C:\\Documents and Settings\\<username>\\Local Settings\\Application Data\\<app_author>\\<app_name>`
        * Win 7  (not roaming):   :bash:`C:\\Users\\<username>\\AppData\\Local\\<app_author>\\<app_name>`
        * Win 7  (roaming):       :bash:`C:\\Users\\<username>\\AppData\\Roaming\\<app_author>\\<app_name>`

    For Unix, we follow the XDG spec and support :bash:`$XDG_DATA_HOME`.
    That means, by default :bash:`~/.local/share/<AppName>`.


    Args:
        str app_name: Name of the application. Will be appended to the base user data path.
        str app_author: Only used in Windows when not in a virtualenv, name of the application author.
        str version: If given, the application version identifier will be appended to the app_name.
        bool roaming: roaming appdata directory. That means that for users on a Windows
                      network setup for roaming profiles, this user data will be synchronized on login. See
                      <http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx> for a discussion of issues.
        bool use_virtualenv: If True and we're running inside of a virtualenv, return a path relative to that
            environment.
        bool create: If True, the folder is created if it does not exist before the path is returned.

    Returns:
        str: the full path to the user data dir for this application.
    """
    return _get_folder(True, _FolderTypes.data, app_name, app_author, version, roaming, use_virtualenv, create)[0]


def user_config_dir(app_name, app_author, version=None, roaming=False, use_virtualenv=True, create=True):
    """
    Return the full path to the user config dir for this application, using a virtualenv location as a base, if it is
    exists, and falling back to the host OS's convention if it doesn't.

    If using a virtualenv, the path returned is :bash:`/path/to/virtualenv/config/app_name`

    Typical user config directories are:
        * Mac OS X:               same as user_data_dir
        * Unix:                   :bash:`~/.config/<AppName> # or in $XDG_CONFIG_HOME, if defined`
        * Win \\*:                  same as user_data_dir

    For Unix, we follow the XDG spec and support $XDG_CONFIG_HOME.
    That means, by default :bash:`~/.config/<AppName>`.

    Args:
        str app_name: Name of the application. Will be appended to the base user config path.
        str app_author: Only used in Windows, name of the application author.
        str version: If given, the application version identifier will be appended to the app_name.
        bool roaming: roaming appdata directory. That means that for users on a Windows
                      network setup for roaming profiles, this user data will be synchronized on login. See
                      <http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx> for a discussion of issues.
        bool use_virtualenv: If True and we're running inside of a virtualenv, return a path relative to that
            environment.
        bool create: If True, the folder is created if it does not exist before the path is returned.

    Returns:
        str: the full path to the user config dir for this application.
    """
    return _get_folder(True, _FolderTypes.config, app_name, app_author, version, roaming, use_virtualenv, create)[0]


def user_cache_dir(app_name, app_author, version=None, use_virtualenv=True, create=True):
    """
    Return the full path to the user cache dir for this application, using a virtualenv location as a base, if it is
    exists, and falling back to the host OS's convention if it doesn't.

    If using a virtualenv, the path returned is :bash:`/path/to/virtualenv/cache/app_name`

    Typical user cache directories are:
        * Mac OS X:   :bash:`~/Library/Caches/<AppName>`
        * Unix:       :bash:`~/.cache/<AppName> (XDG default)`
        * Win XP:     :bash:`C:\\Documents and Settings\\<username>\\Local Settings\\Application Data\\<AppAuthor>\\<AppName>\\Cache`
        * Vista:      :bash:`C:\\Users\\<username>\\AppData\\Local\\<AppAuthor>\\<AppName>\\Cache`

    Args:
        str app_name: Name of the application. Will be appended to the base user config path.
        str app_author: Only used in Windows, name of the application author.
        str version: If given, the application version identifier will be appended to the app_name.
        bool use_virtualenv: If True and we're running inside of a virtualenv, return a path relative to that
            environment.
        bool create: If True, the folder is created if it does not exist before the path is returned.

    Returns:
        str: the full path to the user cache dir for this application.

    """
    return _get_folder(True, _FolderTypes.cache, app_name, app_author, version, False, use_virtualenv, create)[0]


def user_state_dir(app_name, app_author, version=None, roaming=False, use_virtualenv=True, create=True):
    """
    Return the full path to the user state dir for this application, using a virtualenv location as a base, if it is
    exists, and falling back to the host OS's convention if it doesn't.

    If using a virtualenv, the path returned is :bash:`/path/to/virtualenv/state/app_name`

    Typical user state directories are:
        * Mac OS X:  same as user_data_dir
        * Unix:      :bash:`~/.local/state/<AppName>   # or in $XDG_STATE_HOME, if defined`
        * Win \\*:     same as user_data_dir

    For Unix, we follow this Debian proposal https://wiki.debian.org/XDGBaseDirectorySpecification#state
    to extend the XDG spec and support $XDG_STATE_HOME. That means, by default :bash:`~/.local/state/<AppName>`.

    Args:
        str app_name: Name of the application. Will be appended to the base user config path.
        str app_author: Only used in Windows, name of the application author.
        str version: If given, the application version identifier will be appended to the app_name.
        bool roaming: roaming appdata directory. That means that for users on a Windows
                      network setup for roaming profiles, this user data will be synchronized on login. See
                      <http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx> for a discussion of issues.
        bool use_virtualenv: If True and we're running inside of a virtualenv, return a path relative to that
            environment.
        bool create: If True, the folder is created if it does not exist before the path is returned.

    Returns:
        str: the full path to the user state dir for this application.
    """
    return _get_folder(True, _FolderTypes.state, app_name, app_author, version, roaming, use_virtualenv, create)[0]


def user_logs_dir(app_name, app_author, version=None, use_virtualenv=True, create=True):
    """
    Return the full path to the user log dir for this application, using a virtualenv location as a base, if it is
    exists, and falling back to the host OS's convention if it doesn't.

    If using a virtualenv, the path returned is :bash:`:bash:`/path/to/virtualenv/log/app_name``

    Typical user log directories are:
        * Mac OS X:   :bash:`~/Library/Logs/<AppName>`
        * Unix:       :bash:`~/.cache/<AppName>/log  # or under $XDG_CACHE_HOME if defined`
        * Win XP:     :bash:`C:\\Documents and Settings\\<username>\\Local Settings\\Application Data\\<AppAuthor>\\<AppName>\\Logs`
        * Vista:      :bash:`C:\\Users\\<username>\\AppData\\Local\\<AppAuthor>\\<AppName>\\Logs`

    Args:
        str app_name: Name of the application. Will be appended to the base user config path.
        str app_author: Only used in Windows, name of the application author.
        str version: If given, the application version identifier will be appended to the app_name.
        bool use_virtualenv: If True and we're running inside of a virtualenv, return a path relative to that
            environment.
        bool create: If True, the folder is created if it does not exist before the path is returned.

    Returns:
        str: the full path to the user log dir for this application.
    """
    return _get_folder(True, _FolderTypes.logs, app_name, app_author, version, False, use_virtualenv, create)[0]


def site_data_dir(app_name, app_author, version=None, use_virtualenv=True, create=False):
    """
    Return the full path to the OS wide data dir for this application.

    Typical site data directories are:
        * Mac OS X:   :bash:`/Library/Application Support/<AppName>`
        * Unix:       :bash:`/usr/local/share/<AppName> or /usr/share/<AppName>`
        * Win XP:     :bash:`C:\\Documents and Settings\\All Users\\Application Data\\<AppAuthor>\\<AppName>`
        * Vista:      :bash:`(Fail! "C:\\ProgramData" is a hidden *system* directory on Vista.)`
        * Win 7:      :bash:`C:\\ProgramData\\<AppAuthor>\\<AppName>   # Hidden, but writeable on Win 7.`

    For \*nix, this is using the :bash:`$XDG_DATA_DIRS` default.

    .. Note::
        On linux, the $XDG_DATA_DIRS environment variable may contain a list. `site_data_dir` returns the first
        element of this list. If you want access to the whole list, use :func:`site_data_dir_list`

    .. WARNING::
        Do not use this on Windows Vista. See the Vista-Fail note above for why.

    Args:
        str app_name: Name of the application. Will be appended to the base user config path.
        str app_author: Only used in Windows, name of the application author.
        str version: If given, the application version identifier will be appended to the app_name.
        bool use_virtualenv: If True and we're running inside of a virtualenv, return a path relative to that
            environment.
        bool create: If True, the folder is created if it does not exist before the path is returned.

    Returns:
        str: the full path to the site data dir for this application.
    """
    return _get_folder(False, _FolderTypes.data, app_name, app_author, version, False, use_virtualenv, create)[0]


def site_data_dir_list(app_name, app_author, version=None, use_virtualenv=True, create=False):
    """
    Return the list of full path to the OS wide data directories for this application.

    Typical site data directories are:
        * Mac OS X:   :bash:`/Library/Application Support/<AppName>`
        * Unix:       :bash:`/usr/local/share/<AppName> or /usr/share/<AppName>`
        * Win XP:     :bash:`C:\\Documents and Settings\\All Users\\Application Data\\<AppAuthor>\\<AppName>`
        * Vista:      :bash:`(Fail! "C:\\ProgramData" is a hidden *system* directory on Vista.)`
        * Win 7:      :bash:`C:\\ProgramData\\<AppAuthor>\\<AppName>   # Hidden, but writeable on Win 7.`

    For \*nix, this is using the :bash:`$XDG_DATA_DIRS` default.

    .. WARNING::
        Do not use this on Windows Vista. See the Vista-Fail note above for why.

    Args:
        str app_name: Name of the application. Will be appended to the base user config path.
        str app_author: Only used in Windows, name of the application author.
        str version: If given, the application version identifier will be appended to the app_name.
        bool use_virtualenv: If True and we're running inside of a virtualenv, return a path relative to that
            environment.
        bool create: If True, the folder is created if it does not exist before the path is returned.

    Returns:
        list: A list to the full paths for site data directories for this application.
    """
    return _get_folder(False, _FolderTypes.data, app_name, app_author, version, False, use_virtualenv, create)


def site_config_dir(app_name, app_author, version=None, use_virtualenv=True, create=False):
    """
    Return the full path to the OS wide config dir for this application.

    Typical site data directories are:
        * Mac OS X:   :bash:`/Library/Application Support/<AppName>`
        * Unix:       :bash:`/usr/local/share/<AppName> or /usr/share/<AppName>`
        * Win XP:     :bash:`C:\\Documents and Settings\\All Users\\Application Data\\<AppAuthor>\\<AppName>`
        * Vista:      :bash:`(Fail! "C:\\ProgramData" is a hidden *system* directory on Vista.)`
        * Win 7:      :bash:`C:\\ProgramData\\<AppAuthor>\\<AppName>   # Hidden, but writeable on Win 7.`

    For \*nix, this is using the :bash:`$XDG_DATA_DIRS` default.

    .. Note::
        On linux, the $XDG_CONFIG_DIRS environment variable may contain a list. `site_config_dir` returns the first
        element of this list. If you want access to the whole list, use :func:`site_config_dir_list`

    .. WARNING::
        Do not use this on Windows Vista. See the Vista-Fail note above for why.

    Args:
        str app_name: Name of the application. Will be appended to the base user config path.
        str app_author: Only used in Windows, name of the application author.
        str version: If given, the application version identifier will be appended to the app_name.
        bool use_virtualenv: If True and we're running inside of a virtualenv, return a path relative to that
            environment.
        bool create: If True, the folder is created if it does not exist before the path is returned.

    Returns:
        str: the full path to the site config dir for this application.
    """
    return _get_folder(False, _FolderTypes.config, app_name, app_author, version, False, use_virtualenv, create)[0]


def site_config_dir_list(app_name, app_author, version=None, use_virtualenv=True, create=False):
    """
    Return the list of full path to the OS wide data directories for this application.

    Typical site data directories are:
        * Mac OS X:   :bash:`/Library/Application Support/<AppName>`
        * Unix:       :bash:`/usr/local/share/<AppName> or /usr/share/<AppName>`
        * Win XP:     :bash:`C:\\Documents and Settings\\All Users\\Application Data\\<AppAuthor>\\<AppName>`
        * Vista:      :bash:`(Fail! "C:\\ProgramData" is a hidden *system* directory on Vista.)`
        * Win 7:      :bash:`C:\\ProgramData\\<AppAuthor>\\<AppName>   # Hidden, but writeable on Win 7.`

    For \*nix, this is using the :bash:`$XDG_DATA_DIRS` default.

    .. WARNING::
        Do not use this on Windows Vista. See the Vista-Fail note above for why.

    Args:
        str app_name: Name of the application. Will be appended to the base user config path.
        str app_author: Only used in Windows, name of the application author.
        str version: If given, the application version identifier will be appended to the app_name.
        bool use_virtualenv: If True and we're running inside of a virtualenv, return a path relative to that
            environment.
        bool create: If True, the folder is created if it does not exist before the path is returned.

    Returns:
        list: A list to the full paths for site data directories for this application.
    """
    return _get_folder(False, _FolderTypes.config, app_name, app_author, version, False, use_virtualenv, create)


def _get_folder(user, folder_type, app_name, app_author, version, roaming, use_virtualenv, create):
    """
    Get the directory corresponding to the appropriate folder type and operating system.
    The folder is returned, with the app_name, and in the case of windows app_author appended to it.
    If version is not None, it is appended to app_name to allow for multiple versions to run in the same place.

    Since some operating systems have more than one appropriate folder of a given time (e.g. site_data on linux),
    A list is returned. It is up to calling functions to handle the contents of this list.

    Args:
        _FolderTypes folder_type: Folder type value.
        str app_name: Name of the app, the returned dir has os.path.basename == app_name
        str app_author: Name of app author, only used in Windows.
        str version: App version, appended to app_name
        bool roaming: Whether or not the Windows user is roaming.
        bool use_virtualenv: If True and a virtualenv is activated, use the virtualenv path instead of the OS
            convention. Note: This is ignored for site directories, for obvious reasons.
        bool create: If True, create the directory if it doesn't exist. In the case of lists of directories, all folders
            are created.

    Returns:
        list: A list of paths.

    """
    if user and use_virtualenv and _in_virtualenv():
        sub_folder = folder_type.name  # data | config | state | log | cache
        paths = [os.path.join(sys.prefix, sub_folder)]

    elif platform == Platform.WINDOWS:
        if user:
            if folder_type in [_FolderTypes.data, _FolderTypes.config, _FolderTypes.state]:
                paths = [os.path.normpath(_get_win_folder(site=False, roaming=roaming, app_author=app_author))]
            elif folder_type == _FolderTypes.cache:
                # we'll follow the MSDN recommendation on local data, but since they're mum on caches,
                # we'll put them in LOCAL_APPDATA/app_author/Caches.
                path = os.path.normpath(_get_win_folder(site=False, roaming=False, app_author=app_author))
                paths = [os.path.join(path, 'Caches')]
            else:  # folder_type == _FolderTypes.logs:
                # Similar issue as with user caches. MSDN is no help.
                path = os.path.normpath(_get_win_folder(site=False, roaming=False, app_author=app_author))
                paths = [os.path.join(path, 'Logs')]
        elif folder_type in [_FolderTypes.data, _FolderTypes.config]:
            paths = [os.path.normpath(_get_win_folder(site=True, roaming=roaming, app_author=app_author))]
        else:
            raise RuntimeError('Unknown folder type: {}, user: {}'.format(folder_type.name, user))

    elif platform == Platform.MACOS:
        if user:
            if folder_type in [_FolderTypes.data, _FolderTypes.config, _FolderTypes.state]:
                paths = [os.path.expanduser('~/Library/Application Support')]
            elif folder_type == _FolderTypes.cache:
                paths = [os.path.expanduser('~/Library/Caches')]
            else:  # folder_type == _FolderTypes.logs:
                paths = [os.path.expanduser('~/Library/Logs')]
        elif folder_type in [_FolderTypes.data, _FolderTypes.config]:
            paths = [os.path.expanduser('/Library/Application Support')]
        else:
            raise RuntimeError('Unknown folder type: {}, user: {}'.format(folder_type.name, user))

    elif platform == Platform.POSIX:
        if user:
            if folder_type == _FolderTypes.data:
                paths = [os.getenv('XDG_DATA_HOME', os.path.expanduser("~/.local/share"))]
            elif folder_type == _FolderTypes.config:
                paths = [os.getenv('XDG_CONFIG_HOME', os.path.expanduser("~/.config"))]
            elif folder_type == _FolderTypes.state:
                paths = [os.getenv('XDG_STATE_HOME', os.path.expanduser("~/.local/state"))]
            elif folder_type == _FolderTypes.cache:
                paths = [os.getenv('XDG_CACHE_HOME', os.path.expanduser('~/.cache'))]
            else:  # folder_type == _FolderTypes.logs:
                paths = [os.path.expanduser('~/.log')]
        elif folder_type == _FolderTypes.data:
            path = os.getenv('XDG_DATA_DIRS', os.pathsep.join(['/usr/local/share', '/usr/share']))
            paths = [os.path.expanduser(x.rstrip(os.sep)) for x in path.split(os.pathsep)]
        elif folder_type == _FolderTypes.config:
            path = os.getenv('XDG_CONFIG_DIRS', '/etc/xdg')
            paths = [os.path.expanduser(x.rstrip(os.sep)) for x in path.split(os.pathsep)]
        else:
            raise RuntimeError('Unknown folder type: {}, user: {}'.format(folder_type.name, user))

    else:
        raise RuntimeError('Unsupported operating system: {}'.format(sys.platform))

    final_paths = []
    for path in paths:
        final_path = os.path.join(path, app_name)
        if version is not None:
            final_path = '{}_{}'.format(final_path, version)

        if create and not os.path.exists(final_path):
            os.makedirs(final_path)

        final_paths.append(final_path)

    return final_paths


def _in_virtualenv():
    """
    Determine if we're in a virtual env.

    If sys.real_prefix exists, we are in a virtualenv, and sys.prefix is the virtualenv path, while
    sys.real_prefix is the 'system' python.
    if sys.real_prefix does not exist, it could be because we're in python 3 and the user is using
    the built in venv module instead of virtualenv. In this case a sys.base_prefix attribute always exists, and is
    is different from sys.prefix.

    In either case, the path we want in case we are in a virtualenv is sys.prefix.

    Returns:
        bool: True if we are running in a virtual environment, and False otherwise.
    """
    return hasattr(sys, 'real_prefix') or hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix


def _get_win_folder(site, roaming, app_author):
    import ctypes

    if not isinstance(app_author, str) or app_author == '':
        raise RuntimeError('On Windows, app_author must be a non-empty string.')

    # As of Windows Vista, these values have been replaced by KNOWNFOLDERID values.
    # The CSIDL system is supported under Windows Vista, 8 and 10 for compatibility reasons,
    # And the function SHGetFolderPath maps these values to SHGetKnownFolderID. A future version of this
    # library will need to updated if support for CSIDL is ever dropped.
    csidl_consts = {
        "CSIDL_APPDATA": 26,
        "CSIDL_COMMON_APPDATA": 35,
        "CSIDL_LOCAL_APPDATA": 28,
    }

    if site:
        csidl_const = csidl_consts['CSIDL_COMMON_APPDATA']
    elif roaming:
        csidl_const = csidl_consts['CSIDL_APPDATA']
    else:
        csidl_const = csidl_consts['CSIDL_LOCAL_APPDATA']

    buf = ctypes.create_unicode_buffer(1024)
    ctypes.windll.shell32.SHGetFolderPathW(None, csidl_const, None, 0, buf)

    # Downgrade to short path name if have highbit chars. See
    # http://bugs.activestate.com/show_bug.cgi?id=85099.
    # Oren: This bug is not publicly available!
    has_high_char = False
    for c in buf:
        if ord(c) > 255:
            has_high_char = True
            break
    if has_high_char:
        buf2 = ctypes.create_unicode_buffer(1024)
        if ctypes.windll.kernel32.GetShortPathNameW(buf.value, buf2, 1024):
            buf = buf2

    return os.path.join(buf.value, app_author)
